/*
 * Copyright 2017-2020 Dmitry Ivanov
 *
 * This file is part of Quentier.
 *
 * Quentier is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * Quentier is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Quentier. If not, see <http://www.gnu.org/licenses/>.
 */

#include "LogViewerModel.h"
#include "LogViewerModelFileReaderAsync.h"

#include <lib/preferences/keys/Logging.h>

#include <quentier/utility/ApplicationSettings.h>
#include <quentier/utility/DateTime.h>
#include <quentier/utility/EventLoopWithExitStatus.h>
#include <quentier/utility/StandardPaths.h>

#include <QColor>
#include <QCoreApplication>
#include <QFile>
#include <QFileInfo>
#include <QMetaType>
#include <QStringList>
#include <QStringRef>
#include <QTextStream>
#include <QTimeZone>
#include <QTimer>
#include <QTimerEvent>

#include <algorithm>

#define LOG_VIEWER_MODEL_COLUMN_COUNT                (6)
#define LOG_VIEWER_MODEL_NUM_ITEMS_PER_CACHE_BUCKET  (1000)
#define LOG_VIEWER_MODEL_LOG_FILE_POLLING_TIMER_MSEC (500)
#define LOG_VIEWER_MODEL_MAX_LOG_ENTRY_LINE_SIZE     (700)

#define LVMDEBUG(message)                                                      \
    if (m_internalLogEnabled) {                                                \
        QString msg;                                                           \
        QDebug dbg(&msg);                                                      \
        dbg.nospace();                                                         \
        dbg.noquote();                                                         \
        dbg << message;                                                        \
        QString relativeSourceFileName = QStringLiteral(__FILE__);             \
        int prefixIndex = relativeSourceFileName.indexOf(                      \
            QStringLiteral("libquentier"), Qt::CaseInsensitive);               \
        if (prefixIndex >= 0) {                                                \
            relativeSourceFileName.remove(0, prefixIndex);                     \
        }                                                                      \
        else {                                                                 \
            QString appName = QCoreApplication::applicationName().toLower();   \
            prefixIndex =                                                      \
                relativeSourceFileName.indexOf(appName, Qt::CaseInsensitive);  \
            if (prefixIndex >= 0) {                                            \
                relativeSourceFileName.remove(                                 \
                    0, prefixIndex + appName.size() + 1);                      \
            }                                                                  \
        }                                                                      \
        DateTimePrint::Options options(                                        \
            DateTimePrint::IncludeMilliseconds |                               \
            DateTimePrint::IncludeTimezone);                                   \
        QString fullMsg = printableDateTimeFromTimestamp(                      \
                              QDateTime::currentMSecsSinceEpoch(), options) +  \
            QStringLiteral(" ") + relativeSourceFileName +                     \
            QString::fromUtf8(QNLOG_FILE_LINENUMBER_DELIMITER) +               \
            QString::number(__LINE__) + QStringLiteral(": ") + msg +           \
            QStringLiteral("\n");                                              \
        m_internalLogFile.write(fullMsg.toUtf8());                             \
        m_internalLogFile.flush();                                             \
    }                                                                          \
    // LVMDEBUG

namespace quentier {

LogViewerModel::LogViewerModel(QObject * parent) :
    QAbstractTableModel(parent),
    m_internalLogFile(
        applicationPersistentStoragePath() +
        QStringLiteral("/logs-quentier/LogViewerModelLog.txt"))
{
    QObject::connect(
        &m_currentLogFileWatcher, &FileSystemWatcher::fileChanged, this,
        &LogViewerModel::onFileChanged);

    QObject::connect(
        &m_currentLogFileWatcher, &FileSystemWatcher::fileRemoved, this,
        &LogViewerModel::onFileRemoved);

    qRegisterMetaType<QVector<LogViewerModel::Data>>(
        "QVector<LogViewerModel::Data>");

    ApplicationSettings appSettings;
    appSettings.beginGroup(preferences::keys::loggingGroup);

    QVariant enableLogViewerInternalLogsValue =
        appSettings.value(preferences::keys::enableLogViewerInternalLogs);

    appSettings.endGroup();

    bool enableLogViewerInternalLogs = false;
    if (enableLogViewerInternalLogsValue.isValid()) {
        enableLogViewerInternalLogs = enableLogViewerInternalLogsValue.toBool();
    }

    setInternalLogEnabled(enableLogViewerInternalLogs);
}

bool LogViewerModel::isActive() const
{
    return m_isActive;
}

LogViewerModel::~LogViewerModel()
{
    if (m_pFileReaderAsync) {
        m_pFileReaderAsync->disconnect(this);
        m_pFileReaderAsync = nullptr;
    }
}

QString LogViewerModel::logFileName() const
{
    return m_currentLogFileInfo.fileName();
}

void LogViewerModel::setLogFileName(
    const QString & logFileName, const FilteringOptions & filteringOptions)
{
    LVMDEBUG(
        "LogViewerModel::setLogFileName: "
        << logFileName << ", filtering options = " << filteringOptions);

    QString quentierLogFilesDirPath = QuentierLogFilesDirPath();
    QString logFilePath = logFileName;
    if (!logFilePath.startsWith(quentierLogFilesDirPath)) {
        logFilePath =
            quentierLogFilesDirPath + QStringLiteral("/") + logFileName;
    }

    QFileInfo newLogFileInfo(logFilePath);
    if (m_isActive &&
        (m_currentLogFileInfo.absoluteFilePath() ==
         newLogFileInfo.absoluteFilePath()))
    {
        LVMDEBUG("The log file hasn't changed");

        if (m_filteringOptions == filteringOptions) {
            return;
        }
    }

    clear();
    m_currentLogFileInfo = newLogFileInfo;
    m_filteringOptions = filteringOptions;

    QFile currentLogFile(m_currentLogFileInfo.absoluteFilePath());
    if (Q_UNLIKELY(!currentLogFile.exists())) {
        ErrorString errorDescription(QT_TR_NOOP("Log file doesn't exist"));
        errorDescription.details() = m_currentLogFileInfo.absoluteFilePath();
        QNWARNING("model:log_viewer", errorDescription);
        Q_EMIT notifyError(errorDescription);
        return;
    }

    bool open = currentLogFile.open(QIODevice::ReadOnly);
    if (Q_UNLIKELY(!open)) {
        ErrorString errorDescription(
            QT_TR_NOOP("Can't open log file for reading"));
        errorDescription.details() = m_currentLogFileInfo.absoluteFilePath();
        QNWARNING("model:log_viewer", errorDescription);
        Q_EMIT notifyError(errorDescription);
        return;
    }

    m_isActive = true;

    m_currentLogFileStartBytesRead = currentLogFile.read(
        m_currentLogFileStartBytes, sizeof(m_currentLogFileStartBytes));

    currentLogFile.close();

    m_currentLogFileSize = m_currentLogFileInfo.size();
    m_currentLogFileSizePollingTimer.start(
        LOG_VIEWER_MODEL_LOG_FILE_POLLING_TIMER_MSEC, this);

    QString filePath = m_currentLogFileInfo.absoluteFilePath();
    if (!m_currentLogFileWatcher.files().contains(filePath)) {
        m_currentLogFileWatcher.addPath(filePath);
    }

    qint64 startPos =
        (m_filteringOptions.m_startLogFilePos.isSet()
             ? m_filteringOptions.m_startLogFilePos.ref()
             : qint64(0));

    requestDataEntriesChunkFromLogFile(
        startPos, LogFileDataEntryRequestReason::InitialRead);
}

qint64 LogViewerModel::startLogFilePos() const
{
    return m_filteringOptions.m_startLogFilePos.isSet()
        ? m_filteringOptions.m_startLogFilePos.ref()
        : qint64(-1);
}

void LogViewerModel::setStartLogFilePos(const qint64 startLogFilePos)
{
    LVMDEBUG("LogViewerModel::setStartLogFilePos: " << startLogFilePos);

    if (!m_filteringOptions.m_startLogFilePos.isSet() && (startLogFilePos < 0))
    {
        return;
    }

    if (m_filteringOptions.m_startLogFilePos.isSet() &&
        (startLogFilePos >= 0) &&
        (m_filteringOptions.m_startLogFilePos.ref() == startLogFilePos))
    {
        return;
    }

    if (!m_isActive) {
        if (startLogFilePos >= 0) {
            m_filteringOptions.m_startLogFilePos = startLogFilePos;
        }
        else {
            m_filteringOptions.m_startLogFilePos.clear();
        }

        return;
    }

    FilteringOptions filteringOptions = m_filteringOptions;
    if (startLogFilePos >= 0) {
        filteringOptions.m_startLogFilePos = startLogFilePos;
    }
    else {
        filteringOptions.m_startLogFilePos.clear();
    }

    QString logFileName = m_currentLogFileInfo.fileName();
    clear();
    setLogFileName(logFileName, filteringOptions);
}

qint64 LogViewerModel::currentLogFileSize() const
{
    return m_currentLogFileInfo.size();
}

void LogViewerModel::setStartLogFilePosAfterCurrentFileSize()
{
    setStartLogFilePos(m_currentLogFileInfo.size());
}

const QVector<LogLevel> & LogViewerModel::disabledLogLevels() const
{
    return m_filteringOptions.m_disabledLogLevels;
}

void LogViewerModel::setDisabledLogLevels(QVector<LogLevel> disabledLogLevels)
{
    disabledLogLevels.erase(
        std::unique(disabledLogLevels.begin(), disabledLogLevels.end()),
        disabledLogLevels.end());

    if (m_filteringOptions.m_disabledLogLevels.size() ==
        disabledLogLevels.size()) {
        bool foundMismatch = false;
        for (auto it = m_filteringOptions.m_disabledLogLevels.constBegin(),
                  end = m_filteringOptions.m_disabledLogLevels.constEnd();
             it != end; ++it)
        {
            if (!disabledLogLevels.contains(*it)) {
                foundMismatch = true;
                break;
            }
        }

        if (!foundMismatch) {
            return;
        }
    }

    if (!m_isActive) {
        m_filteringOptions.m_disabledLogLevels = disabledLogLevels;
        return;
    }

    FilteringOptions filteringOptions = m_filteringOptions;
    filteringOptions.m_disabledLogLevels = disabledLogLevels;

    QString logFileName = m_currentLogFileInfo.fileName();
    clear();
    setLogFileName(logFileName, filteringOptions);
}

const QString & LogViewerModel::logEntryContentFilter() const
{
    return m_filteringOptions.m_logEntryContentFilter;
}

void LogViewerModel::setLogEntryContentFilter(
    const QString & logEntryContentFilter)
{
    if (m_filteringOptions.m_logEntryContentFilter == logEntryContentFilter) {
        return;
    }

    if (!m_isActive) {
        m_filteringOptions.m_logEntryContentFilter = logEntryContentFilter;
        return;
    }

    FilteringOptions filteringOptions = m_filteringOptions;
    filteringOptions.m_logEntryContentFilter = logEntryContentFilter;

    QString logFileName = m_currentLogFileInfo.fileName();
    clear();
    setLogFileName(logFileName, filteringOptions);
}

bool LogViewerModel::wipeCurrentLogFile(ErrorString & errorDescription)
{
    LVMDEBUG("LogViewerModel::wipeCurrentLogFile");

    if (!m_currentLogFileInfo.exists()) {
        errorDescription.setBase(QT_TR_NOOP("No log file is selected"));
        LVMDEBUG(errorDescription);
        return false;
    }

    if (Q_UNLIKELY(!m_currentLogFileInfo.isFile())) {
        errorDescription.setBase(
            QT_TR_NOOP("Currently selected log file name "
                       "does not really correspond to a file"));
        LVMDEBUG(errorDescription);
        return false;
    }

    if (Q_UNLIKELY(!m_currentLogFileInfo.isWritable())) {
        errorDescription.setBase(
            QT_TR_NOOP("Currently selected log file is not writable"));
        LVMDEBUG(errorDescription);
        return false;
    }

    beginResetModel();

    // NOTE: not stopping the file reader async's thread and not deleting
    // the async file reader immediately, just disconnect from it, mark it for
    // subsequent deletion when possible and lose the pointer to it
    if (m_pFileReaderAsync) {
        m_pFileReaderAsync->disconnect(this);
        m_pFileReaderAsync->deleteLater();
        m_pFileReaderAsync = nullptr;
    }

    QFile currentLogFile(m_currentLogFileInfo.absoluteFilePath());
    bool res = currentLogFile.resize(qint64(0));
    if (Q_UNLIKELY(!res)) {
        errorDescription.setBase(
            QT_TR_NOOP("Failed to clear the contents of the log file"));
        QString errorString = currentLogFile.errorString();
        if (!errorString.isEmpty()) {
            errorDescription.details() = errorString;
        }
        LVMDEBUG(errorDescription);
    }
    else {
        m_currentLogFileSize = 0;

        for (size_t i = 0; i < sizeof(m_currentLogFileStartBytes); ++i) {
            m_currentLogFileStartBytes[i] = 0;
        }
        m_currentLogFileStartBytesRead = 0;

        m_logFileChunksMetadata.clear();
        m_logFileChunkDataCache.clear();
        m_canReadMoreLogFileChunks = false;

        m_logFilePosRequestedToBeRead.clear();
    }

    endResetModel();
    return res;
}

void LogViewerModel::clear()
{
    LVMDEBUG("LogViewerModel::clear");

    beginResetModel();

    m_isActive = false;
    m_currentLogFileInfo = QFileInfo();
    m_currentLogFileWatcher.removePath(m_currentLogFileInfo.absoluteFilePath());

    m_filteringOptions.clear();

    for (size_t i = 0; i < sizeof(m_currentLogFileStartBytes); ++i) {
        m_currentLogFileStartBytes[i] = 0;
    }
    m_currentLogFileStartBytesRead = 0;

    m_logFileChunksMetadata.clear();
    m_logFileChunkDataCache.clear();

    m_canReadMoreLogFileChunks = false;

    m_logFilePosRequestedToBeRead.clear();

    m_currentLogFileSize = 0;
    m_currentLogFileSizePollingTimer.stop();

    // NOTE: not stopping the file reader async's thread and not deleting
    // the async file reader immediately, just disconnect from it, mark it for
    // subsequent deletion when possible and lose the pointer to it
    if (m_pFileReaderAsync) {
        m_pFileReaderAsync->disconnect(this);
        m_pFileReaderAsync->deleteLater();
        m_pFileReaderAsync = nullptr;
    }

    // NOTE: not changing anything about the internal log

    endResetModel();
}

const LogViewerModel::Data * LogViewerModel::dataEntry(const int row) const
{
    int startModelRow = 0;
    const QVector<Data> * pLogFileDataChunk =
        dataChunkContainingModelRow(row, &startModelRow);
    if (!pLogFileDataChunk) {
        return nullptr;
    }

    int offset = row - startModelRow;
    if (Q_UNLIKELY(pLogFileDataChunk->size() <= offset)) {
        return nullptr;
    }

    const Data & data = pLogFileDataChunk->at(offset);
    return &data;
}

const QVector<LogViewerModel::Data> *
LogViewerModel::dataChunkContainingModelRow(
    const int row, int * pStartModelRow) const
{
    if (Q_UNLIKELY(row < 0)) {
        return nullptr;
    }

    const auto * pLogFileChunkMetadata =
        findLogFileChunkMetadataByModelRow(row);

    if (!pLogFileChunkMetadata) {
        return nullptr;
    }

    if (pStartModelRow) {
        *pStartModelRow = pLogFileChunkMetadata->startModelRow();
    }

    return m_logFileChunkDataCache.get(pLogFileChunkMetadata->number());
}

QString LogViewerModel::dataEntryToString(
    const LogViewerModel::Data & dataEntry) const
{
    QString result;
    QTextStream strm(&result);

    strm << dataEntry.m_timestamp.toString(
        QStringLiteral("yyyy-MM-dd HH:mm:ss.zzz t"));

    strm << " " << dataEntry.m_sourceFileName << QNLOG_FILE_LINENUMBER_DELIMITER
         << QString::number(dataEntry.m_sourceFileLineNumber) << " ["
         << LogViewerModel::logLevelToString(dataEntry.m_logLevel)
         << "]: " << dataEntry.m_logEntry;

    strm.flush();
    return result;
}

QColor LogViewerModel::backgroundColorForLogLevel(const LogLevel logLevel) const
{
    int alpha = 140;
    switch (logLevel) {
    case LogLevel::Trace:
        return QColor(127, 174, 255, alpha); // Light blue
    case LogLevel::Debug:
        return QColor(127, 255, 142, alpha); // Light green
    case LogLevel::Info:
        return QColor(252, 255, 127, alpha); // Light yellow
    case LogLevel::Warning:
        return QColor(255, 212, 127, alpha); // Orange
    case LogLevel::Error:
        return QColor(255, 38, 10, alpha); // Red
    default:
        return QColor();
    }
}

void LogViewerModel::saveModelEntriesToFile(const QString & targetFilePath)
{
    LVMDEBUG("LogViewerModel::saveModelEntriesToFile: " << targetFilePath);

    m_targetSaveFile.setFileName(targetFilePath);
    if (!m_targetSaveFile.open(QIODevice::WriteOnly)) {
        ErrorString errorDescription(
            QT_TR_NOOP("Can't save log entries to file: could not open "
                       "the selected file for writing"));
        LVMDEBUG(errorDescription);
        Q_EMIT saveModelEntriesToFileFinished(errorDescription);
        return;
    }

    const LogFileChunksMetadataIndexByNumber & indexByNumber =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByNumber>();
    if (indexByNumber.empty()) {
        LVMDEBUG(
            "The index of log file chunks metadata by number "
            << "is empty, requesting the chunks starting from the first");

        requestDataEntriesChunkFromLogFile(
            0, LogFileDataEntryRequestReason::SaveLogEntriesToFile);

        return;
    }

    qint64 currentLogFileSize = m_currentLogFileInfo.size();

    LVMDEBUG(
        "The number of log file metadata chunks = "
        << indexByNumber.size()
        << ", current log file size = " << currentLogFileSize);

    for (auto it = indexByNumber.begin(), end = indexByNumber.end(); it != end;
         ++it)
    {
        const auto * pDataEntries = m_logFileChunkDataCache.get(it->number());
        if (!pDataEntries) {
            LVMDEBUG(
                "Log data entries are not cached for log "
                << "file metadata chunk with number " << it->number()
                << ", requesting log file data chunks starting from pos "
                << it->startLogFilePos());

            requestDataEntriesChunkFromLogFile(
                it->startLogFilePos(),
                LogFileDataEntryRequestReason::SaveLogEntriesToFile);

            return;
        }

        LVMDEBUG("Processing " << pDataEntries->size() << " data entries");
        for (auto dit = pDataEntries->constBegin(),
                  dend = pDataEntries->constEnd();
             dit != dend; ++dit)
        {
            QString entry = dataEntryToString(*dit);
            entry += QStringLiteral("\n");
            m_targetSaveFile.write(entry.toUtf8());
        }

        if (currentLogFileSize > 0) {
            double progressPercent = static_cast<double>(it->endLogFilePos()) /
                static_cast<double>(currentLogFileSize) * 100.0;

            LVMDEBUG("Emitting save log to file progress: " << progressPercent);
            Q_EMIT saveModelEntriesToFileProgress(progressPercent);
        }
    }

    if (canFetchMore(QModelIndex())) {
        auto lastIt = indexByNumber.end();
        --lastIt;

        LVMDEBUG(
            "The model can fetch more entries, requesting "
            << "data entries starting from pos " << lastIt->endLogFilePos());

        requestDataEntriesChunkFromLogFile(
            lastIt->endLogFilePos(),
            LogFileDataEntryRequestReason::SaveLogEntriesToFile);

        return;
    }

    // If we got here, we found all data entries within the cache, can close
    // the file and return
    LVMDEBUG(
        "Seemingly all log entries were written to the file, "
        << "reporting finish");

    m_targetSaveFile.close();
    Q_EMIT saveModelEntriesToFileFinished(ErrorString());
}

bool LogViewerModel::isSavingModelEntriesToFileInProgress() const
{
    for (auto it = m_logFilePosRequestedToBeRead.constBegin(),
              end = m_logFilePosRequestedToBeRead.constEnd();
         it != end; ++it)
    {
        if (it.value() & LogFileDataEntryRequestReason::SaveLogEntriesToFile) {
            return true;
        }
    }

    return false;
}

void LogViewerModel::cancelSavingModelEntriesToFile()
{
    for (auto it = m_logFilePosRequestedToBeRead.begin();
         it != m_logFilePosRequestedToBeRead.end();)
    {
        if (it.value() ==
            LogFileDataEntryRequestReasons(
                LogFileDataEntryRequestReason::SaveLogEntriesToFile))
        {
            it = m_logFilePosRequestedToBeRead.erase(it);
        }
        else if (
            it.value() & LogFileDataEntryRequestReason::SaveLogEntriesToFile) {
            it.value() &= ~LogFileDataEntryRequestReason::SaveLogEntriesToFile;
            ++it;
        }
    }

    m_targetSaveFile.close();
}

int LogViewerModel::rowCount(const QModelIndex & parent) const
{
    if (parent.isValid()) {
        return 0;
    }

    const auto & indexByNumber =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByNumber>();

    if (indexByNumber.empty()) {
        return 0;
    }

    auto lastIt = indexByNumber.end();
    --lastIt;
    return lastIt->endModelRow() + 1;
}

int LogViewerModel::columnCount(const QModelIndex & parent) const
{
    if (parent.isValid()) {
        return 0;
    }

    return LOG_VIEWER_MODEL_COLUMN_COUNT;
}

QVariant LogViewerModel::data(const QModelIndex & index, int role) const
{
    if (role != Qt::DisplayRole) {
        LVMDEBUG("Not display role (" << role << "), returning empty QVariant");
        return {};
    }

    LVMDEBUG(
        "LogViewerModel::data: index.isValid() = "
        << (index.isValid() ? "true" : "false") << ", index.row() = "
        << index.row() << ", index.column() = " << index.column());

    if (!index.isValid()) {
        LVMDEBUG("Not valid, returning empty QVariant");
        return {};
    }

    int columnIndex = index.column();
    if ((columnIndex < 0) || (columnIndex >= LOG_VIEWER_MODEL_COLUMN_COUNT)) {
        LVMDEBUG("Column outside of valid range, returning empty QVariant");
        return {};
    }

    int rowIndex = index.row();
    if (rowIndex < 0) {
        LVMDEBUG("Row is negative, returning empty QVariant");
        return {};
    }

    const auto * pDataEntry = dataEntry(rowIndex);
    if (pDataEntry) {
        switch (static_cast<Column>(columnIndex)) {
        case Column::Timestamp:
            return pDataEntry->m_timestamp;
        case Column::SourceFileName:
            return pDataEntry->m_sourceFileName;
        case Column::SourceFileLineNumber:
            return pDataEntry->m_sourceFileLineNumber;
        case Column::Component:
            return pDataEntry->m_component;
        case Column::LogLevel:
            return static_cast<qint64>(pDataEntry->m_logLevel);
        case Column::LogEntry:
            return pDataEntry->m_logEntry;
        default:
            return {};
        }
    }

    LVMDEBUG(
        "LogViewerModel::data: data entry was not found for row "
        << rowIndex << " and column " << columnIndex
        << ", returning empty QVariant");

    const auto * pLogFileChunkMetadata =
        findLogFileChunkMetadataByModelRow(rowIndex);

    if (pLogFileChunkMetadata) {
        const_cast<LogViewerModel *>(this)->requestDataEntriesChunkFromLogFile(
            pLogFileChunkMetadata->startLogFilePos(),
            LogFileDataEntryRequestReason::CacheMiss);
    }

    return QVariant();
}

QVariant LogViewerModel::headerData(
    int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole) {
        return QVariant();
    }

    if (orientation == Qt::Vertical) {
        return QVariant(section + 1);
    }

    switch (static_cast<Column>(section)) {
    case Column::Timestamp:
        // TRANSLATOR: log entry's datetime
        return QVariant(tr("Datetime"));
    case Column::SourceFileName:
        // TRANSLATOR: log source file name
        return QVariant(tr("Source file"));
    case Column::SourceFileLineNumber:
        // TRANSLATOR: line number within the source file
        return QVariant(tr("Line number"));
    case Column::Component:
        // TRANSLATOR: logging component
        return QVariant(tr("Component"));
    case Column::LogLevel:
        // TRANSLATOR: the log level i.e. Trace, Debug, Info etc.
        return QVariant(tr("Log level"));
    case Column::LogEntry:
        // TRANSLATOR: the actual recorded log message
        return QVariant(tr("Message"));
    default:
        return QVariant();
    }
}

bool LogViewerModel::canFetchMore(const QModelIndex & parent) const
{
    if (parent.isValid()) {
        return false;
    }

    LVMDEBUG(
        "LogViewerModel::canFetchMore: "
        << (m_canReadMoreLogFileChunks ? "true" : "false"));

    return m_canReadMoreLogFileChunks;
}

void LogViewerModel::fetchMore(const QModelIndex & parent)
{
    if (parent.isValid()) {
        return;
    }

    LVMDEBUG("LogViewerModel::fetchMore");

    qint64 startPos = 0;

    const auto & index =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByStartLogFilePos>();

    if (!index.empty()) {
        auto lastIt = index.end();
        --lastIt;

        startPos = lastIt->endLogFilePos();
    }

    requestDataEntriesChunkFromLogFile(
        startPos, LogFileDataEntryRequestReason::FetchMore);
}

void LogViewerModel::onFileChanged(const QString & path)
{
    if (m_currentLogFileInfo.absoluteFilePath() != path) {
        // The changed file is not the current log file
        return;
    }

    LVMDEBUG("LogViewerModel::onFileChanged");

    m_currentLogFileInfo.refresh();

    QFile currentLogFile(path);
    if (!currentLogFile.isOpen() && !currentLogFile.open(QIODevice::ReadOnly)) {
        ErrorString errorDescription(
            QT_TR_NOOP("Can't open log file for reading"));
        errorDescription.details() = path;
        LVMDEBUG(errorDescription);
        Q_EMIT notifyError(errorDescription);
        return;
    }

    char startBytes[sizeof(m_currentLogFileStartBytes)];
    qint64 startBytesRead = currentLogFile.read(startBytes, sizeof(startBytes));

    currentLogFile.close();

    bool fileStartBytesChanged = false;
    if (startBytesRead != m_currentLogFileStartBytesRead) {
        fileStartBytesChanged = true;
    }
    else {
        for (qint64 i = 0,
                    size = std::min(startBytesRead, qint64(sizeof(startBytes)));
             i < size; ++i)
        {
            size_t index = static_cast<size_t>(i);
            if (startBytes[index] != m_currentLogFileStartBytes[index]) {
                fileStartBytesChanged = true;
                break;
            }
        }
    }

    if (fileStartBytesChanged) {
        // The change within the file is not just the addition of new log entry,
        // the file's start bytes changed, hence should reset the model

        LVMDEBUG(
            "Initial several bytes of the log file have "
            << "changed, the log file was probably rotated");

        m_currentLogFileStartBytesRead = startBytesRead;

        for (qint64 i = 0, size = sizeof(startBytes); i < size; ++i) {
            size_t index = static_cast<size_t>(i);
            m_currentLogFileStartBytes[index] = startBytes[index];
        }

        beginResetModel();

        m_logFileChunksMetadata.clear();
        m_logFileChunkDataCache.clear();
        m_logFilePosRequestedToBeRead.clear();

        m_canReadMoreLogFileChunks = false;

        requestDataEntriesChunkFromLogFile(
            0, LogFileDataEntryRequestReason::InitialRead);

        m_currentLogFileSize = 0;

        endResetModel();

        return;
    }

    // New log entries were appended to the log file, should now be able to
    // fetch more
    LVMDEBUG(
        "The initial bytes of the log file haven't changed "
        "=> new log entries were added, can read more now");

    m_canReadMoreLogFileChunks = true;

    if (m_logFileChunksMetadata.empty()) {
        qint64 startPos =
            (m_filteringOptions.m_startLogFilePos.isSet()
                 ? m_filteringOptions.m_startLogFilePos.ref()
                 : qint64(0));

        requestDataEntriesChunkFromLogFile(
            startPos, LogFileDataEntryRequestReason::InitialRead);
    }
}

void LogViewerModel::onFileRemoved(const QString & path)
{
    if (m_currentLogFileInfo.absoluteFilePath() != path) {
        // The changed file is not the current log file
        return;
    }

    LVMDEBUG("LogViewerModel::onFileRemoved");

    beginResetModel();
    m_currentLogFileInfo = QFileInfo();
    m_currentLogFileWatcher.removePath(path);
    m_currentLogFileStartBytesRead = 0;

    for (size_t i = 0; i < sizeof(m_currentLogFileStartBytes); ++i) {
        m_currentLogFileStartBytes[i] = 0;
    }

    m_logFileChunksMetadata.clear();
    m_logFileChunkDataCache.clear();
    m_logFilePosRequestedToBeRead.clear();

    m_currentLogFileSize = 0;
    m_currentLogFileSizePollingTimer.stop();

    m_canReadMoreLogFileChunks = false;

    endResetModel();
}

void LogViewerModel::onLogFileDataEntriesRead(
    qint64 fromPos, qint64 endPos, QVector<LogViewerModel::Data> dataEntries,
    ErrorString errorDescription)
{
    LVMDEBUG(
        "LogViewerModel::onLogFileDataEntriesRead: from pos = "
        << fromPos << ", end pos = " << endPos << ", num parsed data entries = "
        << dataEntries.size() << ", error description = " << errorDescription);

    auto fromPosIt = m_logFilePosRequestedToBeRead.find(fromPos);
    if (fromPosIt == m_logFilePosRequestedToBeRead.end()) {
        return;
    }

    LogFileDataEntryRequestReasons reasons = fromPosIt.value();
    Q_UNUSED(m_logFilePosRequestedToBeRead.erase(fromPosIt))

    if (!errorDescription.isEmpty()) {
        ErrorString error(
            QT_TR_NOOP("Failed to read a portion of log from file: "));

        error.appendBase(errorDescription.base());
        error.appendBase(errorDescription.additionalBases());
        error.details() = errorDescription.details();

        if (reasons.testFlag(
                LogFileDataEntryRequestReason::SaveLogEntriesToFile)) {
            m_targetSaveFile.close();
            Q_EMIT saveModelEntriesToFileFinished(error);
        }
        else {
            Q_EMIT notifyError(error);
        }

        return;
    }

    if (reasons.testFlag(LogFileDataEntryRequestReason::SaveLogEntriesToFile)) {
        QChar newline(QChar::fromLatin1('\n'));
        for (auto it = dataEntries.constBegin(), end = dataEntries.constEnd();
             it != end; ++it)
        {
            QString entry = dataEntryToString(*it);
            if (!entry.endsWith(newline)) {
                entry += newline;
            }

            m_targetSaveFile.write(entry.toUtf8());
        }

        if (dataEntries.size() < LOG_VIEWER_MODEL_NUM_ITEMS_PER_CACHE_BUCKET) {
            // We must have reached the end of the log file
            m_targetSaveFile.close();
            Q_EMIT saveModelEntriesToFileFinished(ErrorString());
        }
        else {
            qint64 currentLogFileSize = m_currentLogFileInfo.size();
            if (currentLogFileSize > 0) {
                double progressPercent = static_cast<double>(endPos) /
                    static_cast<double>(currentLogFileSize) * 100.0;

                Q_EMIT saveModelEntriesToFileProgress(progressPercent);
            }

            requestDataEntriesChunkFromLogFile(
                endPos, LogFileDataEntryRequestReason::SaveLogEntriesToFile);
        }
    }

    if (reasons ==
        LogFileDataEntryRequestReasons(
            LogFileDataEntryRequestReason::SaveLogEntriesToFile))
    {
        // If there were no other reasons for this request, don't really need to
        // do anything here
        return;
    }

    int logFileChunkNumber = 0;
    int startModelRow = 0;
    int endModelRow = 0;

    if (!dataEntries.isEmpty()) {
        auto & indexByStartPos =
            m_logFileChunksMetadata
                .get<LogFileChunksMetadataByStartLogFilePos>();

        auto it = findLogFileChunkMetadataIteratorByLogFilePos(fromPos);
        bool newEntry = (it == indexByStartPos.end());
        if (newEntry) {
            const auto & indexByNumber =
                m_logFileChunksMetadata.get<LogFileChunksMetadataByNumber>();

            if (!indexByNumber.empty()) {
                auto lastIndexIt = indexByNumber.end();
                --lastIndexIt;
                logFileChunkNumber = lastIndexIt->number() + 1;
                startModelRow = lastIndexIt->endModelRow() + 1;
            }
            else {
                startModelRow = rowCount();
            }

            endModelRow = startModelRow + dataEntries.size() - 1;

            LVMDEBUG(
                "Inserting new rows into the model: start row = "
                << startModelRow << ", end row = " << endModelRow);

            beginInsertRows(QModelIndex(), startModelRow, endModelRow);

            LogFileChunkMetadata metadata(
                logFileChunkNumber, startModelRow, endModelRow, fromPos,
                endPos);

            LVMDEBUG("Appending new log file chunk metadata: " << metadata);
            auto insertResult = m_logFileChunksMetadata.insert(metadata);
            if (!insertResult.second) {
                LVMDEBUG(
                    "The log file chunk metadata was actually "
                    << "updated, not inserted! Here it is: "
                    << *insertResult.first);
            }
        }
        else {
            LogFileChunkMetadata metadata = *it;

            logFileChunkNumber = metadata.number();
            startModelRow = metadata.startModelRow();
            endModelRow = startModelRow + dataEntries.size() - 1;

            metadata = LogFileChunkMetadata(
                logFileChunkNumber, startModelRow, endModelRow, fromPos,
                endPos);

            Q_UNUSED(indexByStartPos.replace(it, metadata))
            LVMDEBUG("Updated log file chunk metadata: " << metadata);
        }

        m_logFileChunkDataCache.put(logFileChunkNumber, dataEntries);
        LVMDEBUG(
            "Put parsed log file data chunk to the LRUCache, "
            << "chunk number = " << logFileChunkNumber);

        if (newEntry) {
            LVMDEBUG("End insert rows");
            endInsertRows();
        }
        else {
            LVMDEBUG(
                "Notifying the view of the model change: "
                << "start row = " << startModelRow
                << ", end row = " << endModelRow);

            auto startIndex =
                index(startModelRow, static_cast<int>(Column::Timestamp), {});

            QModelIndex endIndex =
                index(endModelRow, static_cast<int>(Column::LogEntry), {});

            Q_EMIT dataChanged(startIndex, endIndex);
        }

        LVMDEBUG(
            "Emitting model rows cached signal: start model "
            << "row = " << startModelRow
            << ", end model row = " << endModelRow);

        Q_EMIT notifyModelRowsCached(startModelRow, endModelRow);
    }

    if (dataEntries.size() < LOG_VIEWER_MODEL_NUM_ITEMS_PER_CACHE_BUCKET) {
        // It appears we've read to the end of the log file
        LVMDEBUG("It appears the end of the log file was reached");
        Q_EMIT notifyEndOfLogFileReached();
        m_canReadMoreLogFileChunks = false;
    }
    else {
        LVMDEBUG(
            "Received precisely as many data entries as "
            << "requested, probably more of them can be read");
        m_canReadMoreLogFileChunks = true;
    }
}

void LogViewerModel::requestDataEntriesChunkFromLogFile(
    const qint64 startPos, const LogFileDataEntryRequestReason::type reason)
{
    LVMDEBUG(
        "LogViewerModel::requestDataEntriesChunkFromLogFile: "
        << "start pos = " << startPos << ", request reason = " << reason);

    auto it = m_logFilePosRequestedToBeRead.find(startPos);
    if (it != m_logFilePosRequestedToBeRead.end()) {
        LVMDEBUG(
            "Have already requested log file data entries "
            << "chunk from this start pos, not doing anything");
        it.value() |= reason;
        return;
    }

    if (!m_pReadLogFileIOThread) {
        m_pReadLogFileIOThread = new QThread;

        QObject::connect(
            m_pReadLogFileIOThread, &QThread::finished, this,
            &QThread::deleteLater);

        QObject::connect(
            this, &LogViewerModel::destroyed, m_pReadLogFileIOThread,
            &QThread::quit);

        m_pReadLogFileIOThread->start(QThread::LowPriority);
    }

    if (!m_pFileReaderAsync) {
        m_pFileReaderAsync = new FileReaderAsync(
            m_currentLogFileInfo.absoluteFilePath(),
            m_filteringOptions.m_disabledLogLevels,
            m_filteringOptions.m_logEntryContentFilter);

        m_pFileReaderAsync->moveToThread(m_pReadLogFileIOThread);

        QObject::connect(
            m_pReadLogFileIOThread, &QThread::finished, m_pFileReaderAsync,
            &FileReaderAsync::deleteLater);

        QObject::connect(
            this, &LogViewerModel::readLogFileDataEntries, m_pFileReaderAsync,
            &FileReaderAsync::onReadDataEntriesFromLogFile,
            Qt::ConnectionType(Qt::UniqueConnection | Qt::QueuedConnection));

        QObject::connect(
            m_pFileReaderAsync, &FileReaderAsync::readLogFileDataEntries, this,
            &LogViewerModel::onLogFileDataEntriesRead,
            Qt::ConnectionType(Qt::UniqueConnection | Qt::QueuedConnection));

        QObject::connect(
            this, &LogViewerModel::deleteFileReaderAsync, m_pFileReaderAsync,
            &FileReaderAsync::deleteLater);
    }

    m_logFilePosRequestedToBeRead[startPos] |= reason;

    Q_EMIT readLogFileDataEntries(
        startPos, LOG_VIEWER_MODEL_NUM_ITEMS_PER_CACHE_BUCKET);

    LVMDEBUG(
        "Emitted the request to read no more than "
        << LOG_VIEWER_MODEL_NUM_ITEMS_PER_CACHE_BUCKET
        << " log file data entries starting at pos " << startPos);
}

void LogViewerModel::timerEvent(QTimerEvent * pEvent)
{
    if (Q_UNLIKELY(!pEvent)) {
        return;
    }

    if (pEvent->timerId() == m_currentLogFileSizePollingTimer.timerId()) {
        if (m_currentLogFileInfo.absoluteFilePath().isEmpty()) {
            m_currentLogFileSizePollingTimer.stop();
            return;
        }

        // NOTE: it is necessary to create a new object of QFileInfo type
        // because the existing m_currentLogFileInfo has cached value of
        // current log file size, it doesn't update in live regime
        QFileInfo currentLogFileInfo(m_currentLogFileInfo.absoluteFilePath());
        qint64 size = currentLogFileInfo.size();
        if (size != m_currentLogFileSize) {
            m_currentLogFileSize = size;
            onFileChanged(m_currentLogFileInfo.absoluteFilePath());
        }

        return;
    }

    QAbstractTableModel::timerEvent(pEvent);
}

QString LogViewerModel::logLevelToString(LogLevel logLevel)
{
    switch (logLevel) {
    case LogLevel::Trace:
        return QStringLiteral("Trace");
    case LogLevel::Debug:
        return QStringLiteral("Debug");
    case LogLevel::Info:
        return QStringLiteral("Info");
    case LogLevel::Warning:
        return QStringLiteral("Warn");
    case LogLevel::Error:
        return QStringLiteral("Error");
    default:
        return QStringLiteral("Unknown");
    }
}

void LogViewerModel::setInternalLogEnabled(const bool enabled)
{
    if (m_internalLogEnabled == enabled) {
        return;
    }

    m_internalLogEnabled = enabled;

    if (m_internalLogEnabled) {
        Q_UNUSED(m_internalLogFile.open(QIODevice::WriteOnly))
    }
    else {
        m_internalLogFile.close();
    }
}

bool LogViewerModel::internalLogEnabled() const
{
    return m_internalLogEnabled;
}

const LogViewerModel::LogFileChunkMetadata *
LogViewerModel::findLogFileChunkMetadataByModelRow(const int row) const
{
    const auto & index =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByStartModelRow>();

    auto it = findLogFileChunkMetadataIteratorByModelRow(row);
    if (it == index.end()) {
        return nullptr;
    }

    return &(*it);
}

const LogViewerModel::LogFileChunkMetadata *
LogViewerModel::findLogFileChunkMetadataByLogFilePos(const qint64 pos) const
{
    const auto & index =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByStartLogFilePos>();

    auto it = findLogFileChunkMetadataIteratorByLogFilePos(pos);
    if (it == index.end()) {
        return nullptr;
    }

    return &(*it);
}

LogViewerModel::LogFileChunksMetadataIndexByStartModelRow::const_iterator
LogViewerModel::findLogFileChunkMetadataIteratorByModelRow(const int row) const
{
    const auto & index =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByStartModelRow>();

    auto it = std::lower_bound(
        index.begin(), index.end(), row, LowerBoundByStartModelRowComparator());

    if (it == index.end()) {
        // Unless the index is empty, need to check the previous iterator
        if (index.empty()) {
            return index.end();
        }

        auto lastIt = index.end();
        --lastIt;
        if ((lastIt->startModelRow() < row) && (lastIt->endModelRow() >= row)) {
            return lastIt;
        }

        LVMDEBUG(
            "Appropriate log file chunk metadata was not "
            << "found for model row " << row);

        return index.end();
    }

    if (it->startModelRow() == row) {
        return it;
    }

    if (it == index.begin()) {
        return index.end();
    }

    // By definition of std::lower_bound, it should point to the first metadata
    // entry which has startModelRow greater than or equal to row; if we got
    // here, it is certainly not equal to row but strictly greater. Hence,
    // previous iterator's startModelRow should be less than row, hence previous
    // iterator should point to the metadata entry we are looking for.
    auto prevIt = it;
    --prevIt;
    return prevIt;
}

LogViewerModel::LogFileChunksMetadataIndexByStartLogFilePos::const_iterator
LogViewerModel::findLogFileChunkMetadataIteratorByLogFilePos(
    const qint64 pos) const
{
    const auto & index =
        m_logFileChunksMetadata.get<LogFileChunksMetadataByStartLogFilePos>();

    auto it = std::lower_bound(
        index.begin(), index.end(), pos,
        LowerBoundByStartLogFilePosComparator());

    if (it == index.end()) {
        // Unless the index is empty, need to check the previous iterator
        if (index.empty()) {
            return index.end();
        }

        auto lastIt = index.end();
        --lastIt;
        if ((lastIt->startLogFilePos() < pos) &&
            (lastIt->endLogFilePos() > pos)) {
            return lastIt;
        }

        LVMDEBUG(
            "Log file chunk metadata was not found for log file pos " << pos);

        return index.end();
    }

    if (it->startLogFilePos() == pos) {
        return it;
    }

    if (it == index.begin()) {
        return index.end();
    }

    // By definition of std::lower_bound, it should point to the first metadata
    // entry which has startLogFilePos greater than or equal to pos; if we got
    // here, it is certainly not equal to pos but strictly greater. Hence,
    // previous iterator's startLogFilePos should be less than pos, hence
    // previous iterator should point to the metadata entry we are looking for.
    auto prevIt = it;
    --prevIt;
    return prevIt;
}

LogViewerModel::FilteringOptions::FilteringOptions() = default;

LogViewerModel::FilteringOptions::FilteringOptions(
    const FilteringOptions & filteringOptions) :
    Printable(),
    m_startLogFilePos(filteringOptions.m_startLogFilePos),
    m_disabledLogLevels(filteringOptions.m_disabledLogLevels),
    m_logEntryContentFilter(filteringOptions.m_logEntryContentFilter)
{}

LogViewerModel::FilteringOptions & LogViewerModel::FilteringOptions::operator=(
    const FilteringOptions & filteringOptions)
{
    if (this != &filteringOptions) {
        m_startLogFilePos = filteringOptions.m_startLogFilePos;
        m_disabledLogLevels = filteringOptions.m_disabledLogLevels;
        m_logEntryContentFilter = filteringOptions.m_logEntryContentFilter;
    }

    return *this;
}

LogViewerModel::FilteringOptions::~FilteringOptions() = default;

bool LogViewerModel::FilteringOptions::operator==(
    const FilteringOptions & filteringOptions) const
{
    if (!m_startLogFilePos.isEqual(filteringOptions.m_startLogFilePos)) {
        return false;
    }

    if (m_disabledLogLevels.size() !=
        filteringOptions.m_disabledLogLevels.size()) {
        return false;
    }

    for (auto it = m_disabledLogLevels.constBegin(),
              end = m_disabledLogLevels.constEnd();
         it != end; ++it)
    {
        if (!filteringOptions.m_disabledLogLevels.contains(*it)) {
            return false;
        }
    }

    if (m_logEntryContentFilter != filteringOptions.m_logEntryContentFilter) {
        return false;
    }

    return true;
}

bool LogViewerModel::FilteringOptions::operator!=(
    const FilteringOptions & filteringOptions) const
{
    return !operator==(filteringOptions);
}

bool LogViewerModel::FilteringOptions::isEmpty() const
{
    if (m_startLogFilePos.isSet()) {
        return false;
    }

    if (!m_disabledLogLevels.isEmpty()) {
        return false;
    }

    return m_logEntryContentFilter.isEmpty();
}

void LogViewerModel::FilteringOptions::clear()
{
    m_startLogFilePos.clear();
    m_disabledLogLevels.clear();
    m_logEntryContentFilter.clear();
}

QTextStream & LogViewerModel::FilteringOptions::print(QTextStream & strm) const
{
    strm << "FilteringOptions: start log file pos = ";

    if (m_startLogFilePos.isSet()) {
        strm << m_startLogFilePos.ref();
    }
    else {
        strm << "<not set>";
    }

    strm << ", disabled log levels: ";
    if (!m_disabledLogLevels.isEmpty()) {
        for (int i = 0, size = m_disabledLogLevels.size(); i < size; ++i) {
            switch (m_disabledLogLevels[i]) {
            case LogLevel::Trace:
                strm << "Trace";
                break;
            case LogLevel::Debug:
                strm << "Debug";
                break;
            case LogLevel::Info:
                strm << "Info";
                break;
            case LogLevel::Warning:
                strm << "Warning";
                break;
            case LogLevel::Error:
                strm << "Error";
                break;
            default:
                strm << "Unknown ("
                     << static_cast<qint64>(m_disabledLogLevels[i]) << ")";
                break;
            }

            if (i != (size - 1)) {
                strm << ", ";
            }
        }
    }
    else {
        strm << "<not set>";
    }

    strm << ", log entry content filtering: " << m_logEntryContentFilter;
    return strm;
}

QTextStream & LogViewerModel::Data::print(QTextStream & strm) const
{
    strm << "Timestamp = "
         << printableDateTimeFromTimestamp(m_timestamp.toMSecsSinceEpoch())
         << ", source file name = " << m_sourceFileName
         << ", line number = " << m_sourceFileLineNumber
         << ", component = " << m_component << ", log level = " << m_logLevel
         << ", log entry: " << m_logEntry;
    return strm;
}

QTextStream & LogViewerModel::LogFileChunkMetadata::print(
    QTextStream & strm) const
{
    strm << "Log file chunk: number = " << m_number
         << ", start model row = " << m_startModelRow
         << ", end model row = " << m_endModelRow
         << ", start log file pos = " << m_startLogFilePos
         << ", end log file pos = " << m_endLogFilePos;
    return strm;
}

} // namespace quentier
